#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/* OpenPBS's tags in acct files:
 * 	'Q'	job queued
 * 	'S'	job started
 * 	'E'	job ended
 * 	'D'	job deleted (an entry with tag 'E' will follow)
 * 	'A'	job aborted? (an entry with tag 'E' will follow)
 *
 * Could use the 'start=' and 'end=' entries in an 'E' entry.
 */

#define NEXT_FIELD(p, s) \
	if ( (p=strchr(s, ';')) == NULL) continue; \
	*(p++)='\0';
#define LINE_SIZE	4096

struct JOB {
    int		jobid;
    int		nprocs;
    char	user[32];
    char	group[32];
    long long	stime;
    long long	etime;
    long long	wtime;
    struct JOB	*prev, *next;
};

static struct JOB *cached=NULL, *saved=NULL;
static long long start_time=-1, end_time=-1, period_flag=0;

static long long time2sec(char *time)
/* convert time in the form of "hh:mm:ss" to number of seconds */
{
    int h, m, s;
    long long ll;

    if ( sscanf(time, "%d:%d:%d", &h, &m, &s)==3 ) {
	ll = h*(long long)3600 + m*60 + s;
    } else {
	ll = 0;
    }

    return ll;
}

static long long date2sec(char *time)
/* convert a time string in the form "mm/dd/yyyy hh:mm:ss" to
 * number of seconds since 01/01/1970 00:00:00 */
{
    struct tm t;
    char buff[20];
    time_t t0;

    strncpy(buff, time, 20);
    if ( (buff[2]!='/')  | (buff[5]!='/')  | (buff[10]!=' ') |
	 (buff[13]!=':') | (buff[16]!=':') ) {
	fprintf(stderr, "Invalid time string \"%s\"\n", buff);
	exit(3);
    }
    buff[2]=buff[5]=buff[10]=buff[13]=buff[16]=buff[19]='\0';

    t.tm_isdst = 0;	/* no daylight saving time in our country */
    t.tm_sec   = atoi(buff+17);
    t.tm_min   = atoi(buff+14);
    t.tm_hour  = atoi(buff+11);
    t.tm_mday  = atoi(buff+3);
    t.tm_mon   = atoi(buff+0)-1;
    t.tm_year  = atoi(buff+6)-1900;

    t0=mktime(&t);

    if ( t0 == -1) {
	fprintf(stderr, "date2sec: invalid time string \"%s\"\n", time);
	exit(6);
    }

    return t0;
}

static long long parse_date(const char *date)
/* convert date in the form "yyyymmdd hh:mm:ss" to seconds
 * since 19700101 00:00:00 */
{
    char line[256];

    line[0]=date[4];
    line[1]=date[5];
    line[2]='/';
    line[3]=date[6];
    line[4]=date[7];
    line[5]='/';
    strncpy(line+6, date, 4);

    strcpy(line+10, date+8);
    return date2sec(line);
}

int dates_in_a_month(char *month)
/* return number of dates for the month "yyyymm" */
{
    int y, m;
    char s[256];

    strncpy(s, month+4, 2); s[2]='\0';
    m=atoi(s);
    if ( (m<=7 && (m%2)) || (m>=8 && !(m%2)) ) return 31;
    if ( m!=2 ) return 30;
    strncpy(s, month, 4); s[4]='\0';
    y=atoi(s);
    if ( !(y%100) ) {
	if ( !(y%400) ) return 29; else return 28;
    }
    if ( !(y%4) ) return 29; else return 28;
}

static void parse_period(const char *period)
{
    char start[256], end[256], *p;
    int i;

    if ( !strcmp(period, ".") ) {
	end_time   = parse_date("19900101 00:00:00");
	start_time = parse_date("20291231 23:59:59");
	return;
    }

    period_flag=1;

    p=strchr(period, '-');
    if ( p==NULL ) {
	strcpy(start, period);
	strcpy(end, period);
    } else {
	strncpy(start, period, p-period); start[p-period]='\0';
	strcpy(end, p+1);
    }

    i=strlen(start);
    if ( i!=strlen(end) ) {
	fprintf(stderr, "Start/end date formats do not match: %s\n", period);
	exit(6);
    }

    switch (i) {
	case 4:
	    strcat(start, "01");
	    strcat(end,   "12");
	case 6:
	    strcat(start, "01");
	    sprintf(end+strlen(end), "%02d", dates_in_a_month(end));
	case 8:
	    strcat(start, " 00:00:00");
	    strcat(end,   " 23:59:59");
	    break;
	default:
	    fprintf(stderr, "Invalid period string: %s\n", period);
	    exit(7);
    }

    start_time = parse_date(start);
    end_time   = parse_date(end);
}

char *get_value(char *key, char *desc, int fatal)
{
    char *p, *q=desc;
    static char buffer[LINE_SIZE];
    int n=strlen(key);

    strcpy(buffer, key);
    strcat(buffer, "=");

    do {
	if ( (p=strstr(q, buffer)) == NULL ) {
	    if (fatal) {
		fprintf(stderr, "Cannot match key \"%s\"\n", key);
		fprintf(stderr, "%s", desc);
		exit(5);
	    }
	    buffer[0]='\0';
	    return buffer;
	}
	if ( (p==desc || *(p-1)==' ') ) break;
	q=p+n+1;
    } while (1);

    q=strchr(p+n+1, ' ');
    if ( q==NULL ) strcpy(buffer, p+n+1);
    else {strncpy(buffer, p+n+1, q-p-n-1); buffer[q-p-n-1]='\0';}

#ifdef DEBUG
    fprintf(stderr, "key=%s, value=%s\n", key, buffer);
#endif

    return buffer;
}

static void remove_job(struct JOB *j, struct JOB **root)
{
    if ( j==*root ) *root=j->next;
    if ( j->prev != NULL ) j->prev->next = j->next;
    if ( j->next != NULL ) j->next->prev = j->prev;
}

static void output_job(struct JOB *j)
/* print a job */
{
    char s[1024]="";

    printf( "%-10d %-10s %-8d %-12lld %-12lld%s\n",
		j->jobid,
		j->user,
		j->nprocs,
		j->nprocs * (j->etime - j->stime),
		j->nprocs * j->wtime, s
	  );
}

static struct JOB *new_entry(struct JOB **root)
{
    struct JOB *j;

    /* create a new entry */
    if ( (j=malloc(sizeof(struct JOB))) == NULL ) {
	fprintf(stderr, "Memory allocation error.\n");
	exit(4);
    }
    j->prev=NULL;
    j->next=*root;

    j->stime=-1;
    j->etime=-1;
    j->wtime=0;

    if (*root!=NULL) (*root)->prev=j;

    *root=j;

    return j;
}

static void process_job(char *jobid, char *tag, char *time, char *desc)
{
    struct JOB *j=cached;
    char *p;
    int jid=atoi(jobid);
    long long ll;

    if ( tag[0]!='S' && tag[0]!='E' ) return;

#ifdef DEBUG
    fprintf(stderr, "process_job: jobid=%s, tag=%s, time=%s, desc=%s\n",
		    jobid, tag, time, desc);
#endif

    while (j!=NULL && j->jobid!=jid) j=j->next;

    if ( j==NULL ) {
	j=new_entry(&cached);
	j->jobid=jid;
    }

    ll=date2sec(time);
    if ( !period_flag ) {
	if ( start_time>ll ) start_time=ll;
	if ( end_time<ll ) end_time=ll;
    }
    if ( tag[0]=='S' ) {
	if ( j->stime==-1 || j->stime>ll ) j->stime = ll;
    } else {
	if ( j->etime==-1 || j->etime<ll ) j->etime = ll;
	j->wtime = time2sec(get_value("resources_used.walltime", desc, 0));
    }

    p=get_value("exec_host", desc, 0);
    /* count nbr of '/'s which should equal to nbr of procs */
    j->nprocs=0;
    while ( *p!='\0' ) {
	if ( *p == '/' ) j->nprocs++;
	p++;
    }
    strncpy(j->user,   get_value("user", desc, 1),   sizeof(j->user));
    strncpy(j->group,  get_value("group", desc, 1),  sizeof(j->group));

    if ( j->stime>=0 && j->etime>=0 ) {
	/* move job from 'cached' chain to 'saved' chain */
	remove_job(j, &cached);
	j->next=saved;
	j->prev=NULL;
	if (saved!=NULL) saved->prev=j;
	saved=j;
    }
}

static void do_file(const char *filename)
{
    FILE *f_acct;
    char line[LINE_SIZE], *time, *tag, *jobid, *desc, *p;

    /* process the file */

    if ( (f_acct=fopen(filename,"r")) == NULL ) {
	fprintf(stderr, "Cannot open acct file \"%s\"\n", filename);
	exit(2);
    }

    while ( !feof(f_acct) ) {
	fgets(line, sizeof(line), f_acct);
	/* split the line according to "time;tag;jobid;description" */
	time=line;
	NEXT_FIELD(tag,   time);
	NEXT_FIELD(jobid, tag);
	NEXT_FIELD(desc,  jobid);
	/* remove .s1 (or .lssc) in jobid */
	if ( (p=strchr(jobid, '.')) !=NULL ) *p=0;
	process_job(jobid, tag, time, desc);
    }

    fclose(f_acct);
}

int main(int argc, char *argv[])
{
    int i;
    struct JOB *j, *k, J;
    long long total_time, total_jobs;

    if ( argc < 2+1 ) {
	fprintf(stderr, "Usage: accounting period acct_files\n");
	fprintf(stderr, "Examples of 'period':\n");
	fprintf(stderr, "    . 				all time\n");
	fprintf(stderr, "    2001			year\n");
	fprintf(stderr, "    2001-2003			years\n");
	fprintf(stderr, "    200101			month\n");
	fprintf(stderr, "    200101-200106		months\n");
	fprintf(stderr, "    20010101			day\n");
	fprintf(stderr, "    20010101-20010606		days\n");
	return 1;
    }

    parse_period(argv[1]);

#ifdef DEBUG
    printf("start_time=%lld, end_time=%lld\n", start_time, end_time);
#endif

    fprintf(stderr, "Processing acct files ... ");
    for (i=2; i<argc; i++) do_file(argv[i]);
    fprintf(stderr, "done.\n");

    /* process jobs remained in the cache */
    while (cached!=NULL) {
	j=cached;
	cached=j->next;
	if (j->etime<0) { /* ignore jobs missing etime */
	    free(j);
	    continue;
	}
	if (j->stime<0) j->stime = start_time;
	/* insert j to 'saved' chain */
	j->next=saved;
	j->prev=NULL;
	if (saved!=NULL) saved->prev=j;
	saved=j;
    }

    total_time=total_jobs=0;
    cached=NULL;
    if (saved!=NULL) {
	for (j=saved; j->next!=NULL; j=j->next);
	while (j!=NULL) {
	    if (j->etime>=start_time && j->stime<=end_time) {
		J=*j;
		if (J.stime<start_time) J.stime=start_time;
		if (J.etime>end_time)   J.etime=end_time;
		output_job(&J);
		/* add to per user total: 
		 *	stime=nbr of jobs
		 *	etime=accumulated wall time */
		k=cached;
		while (k!=NULL && strcmp(k->user, J.user)) k=k->next;
		if ( k==NULL ) {
		    k=new_entry(&cached);
		    strcpy(k->user, J.user);
		    k->stime=1;
		    k->etime=(J.etime-J.stime)*J.nprocs;
		} else {
		    (k->stime)++;
		    k->etime+=(J.etime-J.stime)*J.nprocs;
		}
		total_time += (J.etime-J.stime)*J.nprocs;
		total_jobs++;
	    }
	    j=j->prev;
	}
    }

    /* print per user totals */
    if (cached!=NULL) printf("\nPer user statistics: uname seconds jobs\n\n");
    for (j=cached; j!=NULL; j=j->next)
	printf("	%-10s %-12lld %lld\n", j->user, j->etime, j->stime);
    
    printf("\nTotal: %lld seconds, %d%%, %lld jobs\n", 
		    total_time,
		    (int)(total_time/(120.0*(end_time-start_time))*100.0+0.5),
		    total_jobs);

    return 0;
}
