#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#if defined(USE_SIGALRM) || defined(PERFCTR)
#  include <signal.h>
#endif
#include <fcntl.h>
#include <sys/param.h>
#include <asm/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

static void update(int dummy);

#ifdef PERFCTR
extern void do_init(void);
extern int  do_read(double *buffer);
extern int  do_enable(unsigned long sampling_interval);
extern void do_shutdown(void);

static void do_abort(int sig)
{
    do_shutdown();
    exit(1);
}
#endif

static int reset_mflops = 0;
static void do_reset_mflops(int sig)
{
    reset_mflops = 1;
    update(0);
    reset_mflops = 0;
}

static int interval;		/* update every interval milli seconds */
static int perfctr_flag = 1;	/* 0 ==> don't read perf ctrs */
static int nwrite=0;
static int nsensors=0;

static const char *loadinfo_path="/usr/local/loadinfo";
static char path[MAXPATHLEN];
static const char *path0="/usr/spool/PBS/perfctr.dat";
static char buffer[MAXHOSTNAMELEN + 80];

void update(int dummy);

int main(int argc, char *argv[])
{
    char *cp;

    if (argc!=3 && argc!=4) {
	fprintf(stderr, "Usage: loadd update_interval nwrite [nsensors]\n");
	fprintf(stderr, "(read sensors every nsensors*nwrite*update_interval ms"
		        ", <=0 ==> don't read sensors)");
	exit(1);
    }

    interval=atoi(argv[1]);
    if (interval < 0) {
	interval = -interval;
	perfctr_flag = 0;
    }
    nwrite=atoi(argv[2]);
    if (interval <= 0 || nwrite <= 0) {
	fprintf(stderr, "Invalid update interval or nwrite.\n");
	exit(2);
    }

    if (argc==4) nsensors=atoi(argv[3]);

#ifndef DEBUG
    if (getuid()) {
	fprintf(stderr, "loadd: not super user\n");
	exit(3);
    }
#endif

    if (gethostname(buffer, sizeof (buffer) - 1) < 0) {
	fprintf(stderr, "Cannot get hostname.\n");
	exit(1);
    }
    if ((cp = index(buffer, '.')) != NULL) *cp = '\0';
#if 0
    if (!strcmp(buffer, "lssc")) strcpy(buffer, "s1");
#endif
    sprintf(path, "%s/%s", loadinfo_path, buffer);

    fprintf(stderr, "Resource information written to \"%s\".\n", path);

#ifdef PERFCTR
    signal(SIGHUP, do_reset_mflops);
    if (perfctr_flag) {
	atexit(do_shutdown);
	signal(SIGINT, do_abort);
	signal(SIGQUIT, do_abort);
	signal(SIGTERM, do_abort);
	signal(SIGSEGV, do_abort);
	signal(SIGKILL, do_abort);
	do_init();
	if (do_enable(interval * 1000) != 0) {
	   /* try closing and reopening the device */
	   do_shutdown();
	   do_init();
	   do_enable(interval * 1000);
	}
    }
#else
    signal(SIGHUP, SIG_IGN);
#endif

#ifndef DEBUG
    daemon(0, 0);
#endif

#ifndef USE_SIGALRM
    for (;;) {
	struct timeval tv;
	update(0);
	/* usleep(interval * 1000); */
       	tv.tv_sec = interval / 1000;
        tv.tv_usec = (interval % 1000) * 1000;
	select(0, NULL, NULL, NULL, &tv);
    }
#else
#ifdef DEBUG
    fprintf(stderr, "Setting up SIGALRM handler ...\n");
#endif
    signal(SIGALRM, update);
    update(0);
    {
	struct itimerval new, old;
	new.it_value.tv_sec = interval / 1000;
	new.it_value.tv_usec = (interval % 1000) * 1000;
	new.it_interval = new.it_value;
	if (setitimer(ITIMER_REAL, &new, &old) < 0) perror("setitimer");
    }
    while (1) select(0, NULL, NULL, NULL, NULL);
#endif

    return 0;	/* to make gcc happy */
}

static void update(int dummy)
{
    static int locked = 0;
    static double loadavg = 0, cpu_temp = 0, uptime = 0, mbytes = 0;
    static int count0 = -1, count1 = -1;
#ifdef PERFCTR
    static unsigned long mflops_count = 0;
    static double mflops0[]	 = {0, 0, 0, 0};
    static double mflops[]	 = {0, 0, 0, 0};
    static double mflops_total[] = {0, 0, 0, 0};
    static double mflops_max[]	 = {0, 0, 0, 0};
    static int ncpus = 0;
#endif
    FILE *fp;
#ifndef DEBUG
    int fd, size;
    struct stat st;
#endif
    int i;

    if (locked) return;
    locked = 1;

#ifdef PERFCTR
    if (reset_mflops) {
	mflops_count = 0;
	for (i = 0; i < sizeof(mflops) / sizeof(*mflops); i++)
	    mflops_total[i] = mflops_max[i] = 0;
	goto end;
    }
    if (perfctr_flag) {
	ncpus = do_read(mflops0);
	for (i = 0; i < ncpus; i++) {
	    mflops[i] += mflops0[i];
	    mflops_max[i] = (mflops0[i]>mflops_max[i]) ? 
		    		mflops0[i]:mflops_max[i];
	    mflops_total[i] += mflops0[i];
	}
    }
    mflops_count++;

    /* update local file */
    if (ncpus > 0) {
	fd = open(path0, O_WRONLY | O_CREAT, 0644);
	if ( fd > 0) {
	    buffer[0] = '\0';
	    for (i = 0; i < ncpus; i++) {
		sprintf(buffer + strlen(buffer), " %0.2lf", mflops0[i]);
	    }
	}
	/* discard the leading space */
	size=strlen(buffer + 1);
	write(fd, buffer + 1, size);
	if (fstat(fd, &st) < 0 || st.st_size > size)
	ftruncate(fd, size);
	close(fd);
    }
#endif

    if (++count0 < nwrite && count0 > 0) goto end;

    if ( (fp=fopen("/proc/loadavg", "r"))!=NULL ) {
	fscanf(fp, "%lf", &loadavg);
	fclose(fp);
    }

    if ( nsensors > 0 ) {
	if (++count1 <= 0 || count1 >= nsensors) {
	    static struct {
		char *pathname;
		char *format;
	    } sen[]={
		{"/proc/sys/dev/sensors/adm1023-i2c-0-18/temp2",
		 "%*f %*f %*f %lf"},
		{"/proc/sys/dev/sensors/adm1023-i2c-0-1a/temp2",
		 "%*f %*f %*f %lf"},
		{"/proc/sys/dev/sensors/max1617-i2c-0-18/temp2",
		 "%*f %*f %lf"},
		{"/proc/sys/dev/sensors/max1617-i2c-0-1a/temp2",
		 "%*f %*f %lf"},
		{"/proc/sys/dev/sensors/w83627hf-isa-0290/temp2",
		 "%*f %*f %lf"}
	    };
	    cpu_temp = 0;
	    for (i = 0; i < sizeof(sen) / sizeof(sen[0]); i++) {
		if ( (fp=fopen(sen[i].pathname, "r"))!=NULL ) {
		    double t = 0;
		    if (fscanf(fp, sen[i].format, &t)==1 && t>cpu_temp)
			cpu_temp = t;
		    fclose(fp);
		}
	    }
	    count1=0;
	}
    }

    if ( (fp=fopen("/proc/uptime", "r"))!=NULL ) {
	fscanf(fp, "%lf", &uptime);	/* "up_time  idle_time" */
	uptime/=3600;			/* convert to hours */
	fclose(fp);
    }

    if ( (fp=fopen("/proc/meminfo", "r")) != NULL ) {
      mbytes =0.;
      while (1) {
	double m;
	if (fgets(buffer, sizeof(buffer) - 1, fp) == NULL)
	    break;
	if (!memcmp(buffer, "MemTotal:", 9)) {
	    sscanf(buffer, "%*s %lf", &mbytes);
	}
	else if (!memcmp(buffer, "MemFree:", 8)) {
	    sscanf(buffer, "%*s %lf", &m);
	    mbytes -= m;
	}
	else if (!memcmp(buffer, "Buffers:", 8)) {
	    sscanf(buffer, "%*s %lf", &m);
	    mbytes -= m;
	}
	else if (!memcmp(buffer, "Cached:", 7)) {
	    sscanf(buffer, "%*s %lf", &m);
	    mbytes -= m;
	}
#if 0
	else if (!memcmp(buffer, "SwapCached:", 11)) {
	    sscanf(buffer, "%*s %lf", &m);
	    mbytes -= m;
	}
#endif
	else if (!memcmp(buffer, "SwapTotal:", 10)) {
	    sscanf(buffer, "%*s %lf", &m);
	    mbytes += m;
	}
	else if (!memcmp(buffer, "SwapFree:", 9)) {
	    sscanf(buffer, "%*s %lf", &m);
	    mbytes -= m;
	    break;
	}
      }
      fclose(fp);
    }
    mbytes /= 1024.0;

#ifdef DEBUG
    fprintf(stderr, "loadavg: %0.2lf. cpu_temp: %0.1lf, uptime: %0.2lf",
		    loadavg, cpu_temp, uptime);
#ifdef PERFCTR
    if (ncpus > 0) {
	fprintf(stderr, ", mflops:");
	for (i = 0; i < ncpus; i++) {
	    fprintf(stderr, " %0.2lf/%0.2lf/%0.2lf",
			    mflops[i] / (count0 ? nwrite : 1),
			    mflops_total[i] / mflops_count,
			    mflops_max[i]);
	    mflops[i] = 0;
	}
	fprintf(stderr, " %ld", mflops_count);
    }
#endif
    fprintf(stderr, ", mbytes: %0.2lf\n", mbytes);
#else
    /* update global file */
    fd = open(path, O_WRONLY | O_CREAT, 0644);
    if ( fd > 0) {
	sprintf(buffer, "%0.2lf %0.1lf %0.2lf", loadavg, cpu_temp, uptime);
#ifdef PERFCTR
	if (ncpus > 0) {
	    for (i = 0; i < ncpus; i++) {
		sprintf(buffer + strlen(buffer), " %0.2lf/%0.2f/%0.2lf",
				mflops[i] / (count0 ? nwrite : 1),
				mflops_total[i] / mflops_count,
				mflops_max[i]);
		mflops[i] = 0;
	    }
	}
#endif
	sprintf(buffer + strlen(buffer), " %0.2lf", mbytes);
	size=strlen(buffer);
	write(fd, buffer, size);
	if (fstat(fd, &st) < 0 || st.st_size > size)
	ftruncate(fd, size);
	close(fd);
    }
#endif

    count0 = 0;

end:
    locked = 0;
}
