/* $Id: global.c,v 1.1.1.1 2003/10/11 16:13:30 mucci Exp $
 * global.c
 *
 * usage: ./global [sampling_interval_usec [sleep_interval_sec]]
 *
 * This test program illustrates how a process may use the
 * Linux x86 Performance-Monitoring Counters interface to
 * do system-wide performance monitoring.
 *
 * Copyright (C) 2000-2002  Mikael Pettersson
 */
#include <errno.h>
#include <setjmp.h>
#include <signal.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "libperfctr.h"

struct gperfctr *gperfctr = NULL;
struct perfctr_info info;
#define have_rdtsc (info.cpu_features & PERFCTR_FEATURE_RDTSC)
struct gperfctr_control *control = NULL;
struct gperfctr_state *state = NULL;
struct gperfctr_state *prev_state = NULL;
int counting_mips = 0;	/* default is to count MFLOP per second */

static unsigned long sampling_interval_save = 1000000;

int do_enable(unsigned long sampling_interval);

#ifdef TEST
jmp_buf main_buf;

void onint(int sig)	/* ^C handler */
{
    longjmp(main_buf, 1);
}
#endif

void do_shutdown(void)
{
    if ( gperfctr ) {
	printf("shutting down..\n");
	gperfctr_stop(gperfctr);
	gperfctr = NULL;
    }
}

void
do_init(void)
{
#ifdef TEST
    struct sigaction act;
#endif
    size_t nbytes;

    gperfctr = gperfctr_open();
    if( !gperfctr ) {
	perror("gperfctr_open");
	return;
    }
    if( gperfctr_info(gperfctr, &info) < 0 ) {
	perror("gperfctr_info");
	do_shutdown();
	return;
    }
#ifdef TEST
    printf("\nPerfCtr Info:\n");
    perfctr_print_info(&info);

    memset(&act, 0, sizeof act);
    act.sa_handler = onint;
    if( sigaction(SIGINT, &act, NULL) < 0 ) {
	perror("unable to catch SIGINT");
	do_shutdown();
	return;
    }
#endif

    /* now alloc control and state memory based on nrcpus */

    nbytes = offsetof(struct gperfctr_control, cpu_control[0])
	+ info.nrcpus * sizeof(control->cpu_control[0]);
    control = realloc(control, nbytes);
    if( !control ) {
	perror("malloc");
	do_shutdown();
	return;
    }
    memset(control, 0, nbytes);

    nbytes = offsetof(struct gperfctr_state, cpu_state[0])
	+ info.nrcpus * sizeof(state->cpu_state[0]);
    state = realloc(state, nbytes);
    prev_state = realloc(prev_state, nbytes);
    if( !state || !prev_state ) {
	perror("malloc");
	do_shutdown();
	return;
    }
    memset(state, 0, nbytes);
    memset(prev_state, 0, nbytes);
}

int do_read(double *buffer)
{
    int cpu, nactive;

    if (gperfctr == NULL) return 0;

    state->nrcpus = info.nrcpus;
    if( (nactive = gperfctr_read(gperfctr, state)) < 0 ) {
	/* try to reinitialize */
	do_shutdown();
	do_init();
	do_enable(sampling_interval_save);
	if( (nactive = gperfctr_read(gperfctr, state)) < 0 ) {
	    perror("gperfctr_read");
	    do_shutdown();
	    return 0;
	}
    }
    for(cpu = 0; cpu < nactive; ++cpu) {
	if( state->cpu_state[cpu].cpu_control.nractrs >= 1 ) {	/* compute and display MFLOP/s or MIP/s */
	    unsigned long long tsc = state->cpu_state[cpu].sum.tsc;
	    unsigned long long prev_tsc = prev_state->cpu_state[cpu].sum.tsc;
	    unsigned long long ticks = tsc - prev_tsc;
	    unsigned long long pmc0 = state->cpu_state[cpu].sum.pmc[0];
	    unsigned long long prev_pmc0 = prev_state->cpu_state[cpu].sum.pmc[0];
	    unsigned long long ops = pmc0 - prev_pmc0;
	    double seconds = have_rdtsc
		? ((double)ticks / (double)info.cpu_khz) / 1000.0
		: (double)1.0; /* don't div-by-0 on WinChip ... */
	    buffer[cpu] = ((double)ops / (seconds ? : 1)) / 1e6;
	    prev_state->cpu_state[cpu].sum.tsc = tsc;
	    prev_state->cpu_state[cpu].sum.pmc[0] = pmc0;
	}
    }
    return nactive;
}

void setup_control(struct perfctr_cpu_control *control)
{
    unsigned int tsc_on = 1;
    unsigned int nractrs = 1;
    unsigned int pmc_map0 = 0;
    unsigned int evntsel0 = 0;
    unsigned int evntsel_aux0 = 0;

    if (gperfctr == NULL) return;

    /* Attempt to set up control to count clocks via the TSC
       and FLOPS via PMC0. */
    switch( info.cpu_type ) {
      case PERFCTR_X86_GENERIC:
	nractrs = 0;		/* no PMCs available */
	break;
      case PERFCTR_X86_INTEL_P5:
      case PERFCTR_X86_INTEL_P5MMX:
      case PERFCTR_X86_CYRIX_MII:
	/* event 0x22 (FLOPS), any CPL */
	evntsel0 = 0x22 | (3 << 6);
	break;
      case PERFCTR_X86_INTEL_P6:
      case PERFCTR_X86_INTEL_PII:
      case PERFCTR_X86_INTEL_PIII:
	/* note: FLOPS is only available in PERFCTR0 */
	/* event 0xC1 (FLOPS), any CPL, Enable */
	evntsel0 = 0xC1 | (3 << 16) | (1 << 22);
	break;
      case PERFCTR_X86_AMD_K7:
	/* K7 apparently can't count FLOPS. */
	counting_mips = 1;
	/* event 0xC0 (RETIRED_INSTRUCTIONS), any CPL, Enable */
	evntsel0 = 0xC0 | (3 << 16) | (1 << 22);
	break;
      case PERFCTR_X86_WINCHIP_C6:
	counting_mips = 1;	/* can't count FLOPS */
	tsc_on = 0;		/* no working TSC available */
	evntsel0 = 0x02;	/* X86_INSTRUCTIONS */
	break;
      case PERFCTR_X86_WINCHIP_2:
	counting_mips = 1;	/* can't count FLOPS */
	tsc_on = 0;		/* no working TSC available */
	evntsel0 = 0x16;	/* INSTRUCTIONS_EXECUTED */
	break;
      case PERFCTR_X86_VIA_C3:
	counting_mips = 1;	/* can't count FLOPS */
	pmc_map0 = 1;		/* redirect PMC0 to PERFCTR1 */
	evntsel0 = 0xC0;	/* INSTRUCTIONS_EXECUTED */
	break;
      case PERFCTR_X86_INTEL_P4:
      case PERFCTR_X86_INTEL_P4M2:
	nractrs = 2;
	/* set up PMC(1) to produce tagged x87_FP_uop:s */
	control->pmc_map[1] = 0x8 | (1 << 31);
	control->evntsel[1] = (0x3 << 16) | (1 << 13) | (1 << 12);
	control->evntsel_aux[1] = (4 << 25) | (1 << 24) | (1 << 5) | (1 << 4) | (1 << 2);
	/* set up PMC(0) to count execution_event(X87_FP_retired) */
	pmc_map0 = 0xC | (1 << 31);
	evntsel0 = (0x3 << 16) | (5 << 13) | (1 << 12);
	evntsel_aux0 = (0xC << 25) | (1 << 9) | (1 << 2);
	break;
      default:
	fprintf(stderr, "cpu_type %u (%s) not supported\n",
		info.cpu_type, perfctr_cpu_name(&info));
	do_shutdown();
	return;
    }
    control->tsc_on = tsc_on;
    control->nractrs = nractrs;
    control->pmc_map[0] = pmc_map0;
    control->evntsel[0] = evntsel0;
    control->evntsel_aux[0] = evntsel_aux0;
}

int do_enable(unsigned long sampling_interval)
{
    int cpu;

    if (gperfctr == NULL) return 1;

    setup_control(&control->cpu_control[0]);

    for(cpu = 1; cpu < info.nrcpus; ++cpu)
	control->cpu_control[cpu] = control->cpu_control[0];
    sampling_interval_save = control->interval_usec = sampling_interval;
    control->nrcpus = info.nrcpus;		/* use all available cpus */

    if( gperfctr_control(gperfctr, control) < 0 ) {
	perror("gperfctr_control");
	do_shutdown();
	return 1;
    }

    return 0;
}

#if TEST
static unsigned long sampling_interval = 1000000; /* reduce for >4GHz CPUs */
static unsigned int sleep_interval = 5;

int main(int argc, const char **argv)
{
    if( argc >= 2 ) {
	sampling_interval = strtoul(argv[1], NULL, 0);
	if( argc >= 3 )
	    sleep_interval = strtoul(argv[2], NULL, 0);
    }

    do_init();
    do_enable(sampling_interval);

    printf("\nSampling interval:\t%lu usec\n", sampling_interval);
    printf("Sleep interval:\t\t%u sec\n", sleep_interval);

    if( setjmp(main_buf) == 0 ) {
	do {
	    sleep(sleep_interval);
	} while( do_read() == 0 );
    }
    printf("shutting down..\n");
    gperfctr_stop(gperfctr);
    return 0;
}
#endif
